# -*- coding: utf-8 -*-
from django.test import TestCase

from pyoneall import OneAll


class TestOneAll(TestCase):
    def setUp(self):
        self._oa = OneAll('py3tests', 'bf3a6a88-5300-4880-8982-acabe89f9cd1', '35fc1a5e-fa81-4297-a0e6-723bcca838ee')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import uuid
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='SocialUserCache',
            fields=[
                ('user_token', models.UUIDField(serialize=False, primary_key=True)),
                ('raw', models.TextField(default='{}')),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='EmailLoginToken',
            fields=[
                ('token', models.UUIDField(serialize=False, default=uuid.uuid4, primary_key=True)),
                ('email', models.EmailField(max_length=254, unique=True)),
                ('created', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from base64 import urlsafe_b64decode, urlsafe_b64encode
from logging import getLogger
from uuid import UUID

from django.contrib.auth.backends import ModelBackend
from django.http import QueryDict
from pyoneall import OneAll

from .app import settings
from .models import SocialUserCache, EmailLoginToken, get_pseudo_random_user

log = getLogger(__name__)

# The worker to be used for authentication
oneall = OneAll(**settings.credentials)


class BaseBackend(object):
    def __init__(self, existing_user=None):
        self.user = existing_user

    @classmethod
    def get_user(cls, user_id):
        """
        Retrieve user by user ID
        :param user_id: User ID
        """
        return ModelBackend().get_user(user_id)


class OneAllAuthBackend(BaseBackend):
    KEY = 'connection_token'

    def authenticate(self, connection_token, **_):
        """
        Performs authentication using a connection token. Creates and updates User and OneAllUserIdentity
        if necessary.
        :param str connection_token: OneAll connection token
        """
        oa_user = oneall.connection(connection_token).user

        if not settings.store_user_info:
            # Do not store any personally identifiable information.
            return get_pseudo_random_user(UUID(oa_user.user_token))

        # Check if user exists and create one if not
        try:
            identity = SocialUserCache.objects.get(user_token=oa_user.user_token)
            identity.refresh(raw=oa_user.identity)
            if self.user:
                identity.user = self.user  # override any existing link.
            identity.update_user_cache()
        except SocialUserCache.DoesNotExist:
            identity = SocialUserCache(user_token=oa_user.user_token,
                                       raw=str(oa_user.identity), user=self.user)
            identity.update_user_cache()

        # Return authenticated user
        return identity.user


class EmailTokenAuthBackend(BaseBackend):
    KEY = 'etk'
    login = None

    def issue(self, email):
        self.login = EmailLoginToken.issue(email)
        result = QueryDict(mutable=True)
        result[self.KEY] = urlsafe_b64encode(self.login.token.bytes)[:-2]
        return result

    def authenticate(self, **kwargs):
        if self.KEY in kwargs:
            token = UUID(bytes=urlsafe_b64decode(kwargs[self.KEY].encode('ascii') + b'=='))
            try:
                login = EmailLoginToken.consume(token)
            except EmailLoginToken.DoesNotExist:
                return None
            if not self.user:
                return login.produce_user()
            else:
                # this was an email change request!
                # TODO: check for existing users bearing the same address. can a merge happen? if not, what can?
                self.user.email = login.email
                self.user.save()
                return self.user



# -*- coding: utf-8 -*-
from django.conf.urls import url

from . import views

urlpatterns = [
    url('^login$', views.oa_login, name='oneall-login'),
    url('^login(.+)$', views.oa_login, name='oneall-login'),
    url('^logout', views.oa_logout, name='oneall-logout'),
    url('^profile', views.oa_profile, name='oneall-profile'),
]



# -*- coding: utf-8 -*-
from logging import getLogger
from random import Random
from re import match, sub
from uuid import uuid4

from django.contrib.auth.models import User
from django.db import models
from django.db.transaction import atomic
from django.utils.timezone import now
from pyoneall.base import OADict

from .app import settings

log = getLogger(__name__)

USERNAME_MAXLEN = User._meta.get_field('username').max_length


class SocialUserCache(models.Model):
    """
    OneAll User Identity Model
    Caches raw JSON corresponding with user's social identity allow instant retrieval of user details.
    """
    user_token = models.UUIDField(primary_key=True)
    raw = models.TextField(default='{}')
    user = models.ForeignKey(User, null=True)

    def __init__(self, *args, **kwargs):
        """
        Upon creation, creates attributes to correspond with the cached data in `raw`
        """
        super(self.__class__, self).__init__(*args, **kwargs)
        self.__dict__.update(OADict(**eval(self.raw)))

    def refresh(self, raw=None):
        """
        Refresh identity cache from OneAll
        """
        if not raw:
            from .auth import oneall

            raw = oneall.user(self.user.username).identities.identity[0]
        raw.pop('id', None)
        raw.pop('user', None)
        self.raw = str(raw)
        self.__dict__.update(OADict(**eval(self.raw)))
        self.save()

    def update_user_cache(self):
        """
        Update selected fields in the User model from social identity
        """
        user = self.user
        if 'emails' in self.__dict__ and self.emails:
            email = self.emails[0].value
            if not user:
                user = User.objects.filter(email=email).first() or User(email=email)
            else:
                user.email = email
        if not user:
            user = User()
        if 'name' in self.__dict__ and 'givenName' in self.name:
            user.first_name = self.name.givenName
        if 'name' in self.__dict__ and 'familyName' in self.name:
            user.last_name = self.name.familyName
        if not user.username:
            username = getattr(self, 'preferredUsername', None)
            username = username or sub(r'@.*$', '', str(self.emails[0].value))
            username = _find_unique_username(username)
            user.username = username
        user.save()
        self.user = user
        self.save()


class EmailLoginToken(models.Model):
    token = models.UUIDField(primary_key=True, default=uuid4)
    email = models.EmailField(unique=True)
    created = models.DateTimeField(auto_now=True)

    @classmethod
    @atomic
    def issue(cls, email):
        cls._expire()
        cls.objects.filter(email=email).delete()
        login = cls(email=email)
        login.save()
        return login

    @classmethod
    @atomic
    def consume(cls, token):
        cls._expire()
        login = cls.objects.get(token=token)
        login.delete()
        return login  # Despite being removed from the db, we're still going to use this object for a bit longer.

    def produce_user(self):
        user = User.objects.filter(email=self.email).first()
        if not user:
            preferred_username = sub(r'@.*$', '', str(self.email))
            user = User(email=self.email, username=_find_unique_username(preferred_username))
            user.save()
        return user

    @classmethod
    @atomic
    def _expire(cls):
        cls.objects.filter(created__lt=now() - settings.token_expiration).delete()


def _find_unique_username(current):
    """
    Checks wether given username is unique.
    If not unique or not given, tries to derive a new username that is.
    """

    def exists(n):
        return User.objects.filter(username=n).exists()

    def cat_maxlen(a, b, l):
        return a[:l - len(b) - 1] + b

    current = str(current)[:USERNAME_MAXLEN]
    if current and not exists(current):
        return current
    prefix, suffix = match(r'^(.+?)(\d*)$', current or 'user').groups()
    suffix = int(suffix or 0) + 1
    current = cat_maxlen(prefix, str(suffix), USERNAME_MAXLEN)
    while exists(current):
        suffix += 1
        current = cat_maxlen(prefix, str(suffix), USERNAME_MAXLEN)
    return current


def get_pseudo_random_user(seed):
    """ Creates or finds a user with a pseudo-random username and no further information.
    :param seed: A hashable object used to find the user. Will not be stored. """
    alphabet = 'AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz'  # [chr(i+65)+chr(i+97) for i in range(26)]
    not_so_random = Random(seed)
    username = ''.join(map(lambda x: not_so_random.choice(alphabet), range(USERNAME_MAXLEN)))
    try:
        return User.objects.get(username=username)
    except User.DoesNotExist:
        user = User(username=username)
        user.save()
        return user



# -*- coding: utf-8 -*-
from uuid import UUID

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.core.urlresolvers import reverse

from ...auth import EmailTokenAuthBackend
from ...models import SocialUserCache


class Command(BaseCommand):
    help = "Sets that user as super user. To create it too, use e-mail login."

    def add_arguments(self, parser):
        parser.add_argument('user', type=str, help="User ID, e-mail, or OneAll identity UUID.")

    def _extract_user(self, user):
        user_model = get_user_model()
        try:
            return user_model.objects.get(id=int(user))
        except (ValueError, user_model.DoesNotExist):
            pass
        try:
            return SocialUserCache.objects.get(user_token=UUID(user)).user
        except (ValueError, SocialUserCache.DoesNotExist):
            pass
        if '@' in user:
            auth = EmailTokenAuthBackend()
            self.stdout.write("Login with: %s?%s" % (reverse('oneall-login'), auth.issue(user)))
            return auth.login.produce_user()
        self.stdout.write("User <%s> not found." % user)

    def handle(self, user, **options):
        user = self._extract_user(user)
        if not user:
            return
        user.is_active = True
        user.is_staff = True
        user.is_superuser = True
        user.save()
        self.stdout.write("User #%d, email=%s is now super user." % (user.id, user.email))



# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import SocialUserCache


@admin.register(SocialUserCache)
class IdentityAdmin(admin.ModelAdmin):
    readonly_fields = ('raw',)
    list_display = ('user_token', 'user', 'raw')



# -*- coding: utf-8 -*-
from smtplib import SMTPResponseException

from django.core.mail import EmailMessage
from django.core.management.base import BaseCommand
from django.core.urlresolvers import reverse

from ...auth import EmailTokenAuthBackend


class Command(BaseCommand):
    help = "Issues an e-mail login token."

    def add_arguments(self, parser):
        parser.add_argument('-s', '--send', dest='send', action='store_true',
                            help="Actually e-mail the token instead of only displaying it.")
        parser.add_argument('email', type=str)

    def handle(self, email, send, **options):
        if '@' not in email:
            self.stderr.write("Failed. E-mail is mandatory.")
            return
        query_string = EmailTokenAuthBackend().issue(email)
        msg = "Complete login with: %s?%s" % (reverse('oneall-login'), query_string)
        self.stdout.write(msg)
        if send:
            mail = EmailMessage()
            mail.to = [email]
            mail.subject = 'Login Test'
            mail.body = msg
            try:
                sent = mail.send()
                self.stdout.write("Sent %d message." % sent)
            except SMTPResponseException as e:
                self.stderr.write("SMTP %d %s" % (e.smtp_code, e.smtp_error.decode('latin9')))
            except OSError as e:
                self.stderr.write("Conn %s" % e)



# -*- coding: utf-8 -*-
from json import dumps

from django.template import Library
from django.utils.safestring import mark_safe

from ..app import settings
from ..models import User, SocialUserCache

register = Library()


@register.inclusion_tag('oneall/header.html')
def oneall_header():
    """
    OneAll required script.

    This must go in the ``<head>...</head>`` section of your templates,
    otherwise widgets won't load.
    """
    return {'oneall_site_name': settings.credentials['site_name']}


@register.inclusion_tag('oneall/social_login.html')
def oneall_social_login(user=None, **kwargs):
    """
    This tag displays the Social Login or Social Link widget.

    Don't forget to include ``{% oneall_header %}``!

    :param user: Logged in user for Social Link mode; if not provided, it's Social Login mode.
    :param kwargs: Widget options as documented by OneAll. For example, ``grid_sizes=[8,5]``
    """
    if isinstance(user, User):
        social_user = SocialUserCache.objects.filter(user=user).first()
        if social_user:
            kwargs['user_token'] = str(social_user.user_token)
        else:
            user = None  # no cached social user, thus revert to social login mode
    widget_settings = {}
    for key, value in settings.login_widget(kwargs).items():
        widget_settings[key] = mark_safe(dumps(value))
    return {
        'settings': widget_settings,
        'mode': 'social_link' if user else 'social_login',
    }



# -*- coding: utf-8 -*-
from os.path import join, dirname

from setuptools import setup, find_packages

README = open(join(dirname(__file__), 'README.rst')).read()
setup(
    name='django-oneall',
    version='1.0',
    packages=find_packages(),
    package_data={
        'django_oneall/templates/oneall': [
            'header.html',
            'login.html',
            'profile.html',
            'social_login.html'
        ],
    },
    include_package_data=True,
    install_requires=['pyoneall>=0.2.1', 'django>=1.8'],
    license='MIT License, see LICENSE file',
    description='Django Authentication support for OneAll. Provides unified authentication for 30+ social networks',
    long_description=README,
    url='http://www.leandigo.com/',
    author='Michael Greenberg / Leandigo',
    author_email='michael@leandigo.com',
    maintainer='Ekevoo',
    maintainer_email='ekevoo@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django :: 1.8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        # Python 3.2 is not supported because "future" is incompatible with it.
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
    ],
)



# -*- coding: utf-8 -*-
from uuid import UUID

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import models, OperationalError, connection

from ...models import SocialUserCache


class Command(BaseCommand):
    help = "Loads OneAll data from a legacy table (oneall_cache table from django-oneall v0.1)."

    class LegacyOneAllCache(models.Model):
        user_token = models.CharField(max_length=36, primary_key=True)
        raw = models.CharField(max_length=8192)
        user = models.ForeignKey(to=settings.AUTH_USER_MODEL)

        class Meta(object):
            db_table = 'oneall_cache'

    def handle(self, *args, **options):
        try:
            src = self.LegacyOneAllCache.objects
            self.stdout.write("%d users to migrate." % src.count())
        except OperationalError as e:
            self.stderr.write("Couldn't import. %s" % e)
            return
        cok = cbad = 0
        for oldrow in src.iterator():
            try:
                newid = UUID(oldrow.user_token)
                newrow = SocialUserCache(user_token=newid, raw=oldrow.raw, user=oldrow.user)
                newrow.save()
                cok += 1
            except ValueError:
                cbad += 1
                self.stderr.write("Invalid UUID <%s>", oldrow.user_token)
        self.stdout.write("Done. %d ok, %d failed." % (cok, cbad))
        if cbad:
            self.stdout.write("Import complete with errors. Keeping legacy table.")
        else:
            self.stdout.write("Import complete without errors. Deleting legacy table.")
            with connection.schema_editor() as schema_editor:
                schema_editor.delete_model(self.LegacyOneAllCache)



# -*- coding: utf-8 -*-
from datetime import timedelta

from django.conf import settings as django_settings


class MissingOneAllSettings(KeyError):
    def __init__(self, msg=None):
        msg = 'Missing or invalid settings. Check django-oneall documentation for further info. ' + (msg or '')
        super().__init__(msg)


class AppSettings(object):
    def __init__(self):
        if not hasattr(django_settings, 'ONEALL'):
            raise MissingOneAllSettings
        self._settings = django_settings.ONEALL
        if not isinstance(self._settings, dict):
            raise MissingOneAllSettings
        self._default_login_widget = {
            None: "This is a dummy marker to indicate that the Django settings haven't been loaded yet.",
            'providers': 'google facebook twitter openid'.split(),
        }

    @property
    def credentials(self):
        """ Provides the credentials dictionary for the OneAll connection. """
        try:
            return self._settings['credentials']
        except KeyError:
            raise MissingOneAllSettings('OneAll site credentials missing.')

    def login_widget(self, overlay=None):
        """ Provides the login widget settings as a dictionary, with an optional overlay. """
        if None in self._default_login_widget:
            del self._default_login_widget[None]
            self._default_login_widget.update(self._settings.get('login_widget', {}))
        if overlay:
            result = dict(self._default_login_widget)
            result.update(overlay)
            return result
        else:
            return self._default_login_widget

    @property
    def store_user_info(self):
        """ Whether to store personally identifiable information from users. Defaults to True. """
        return bool(self._settings.get('store_user_info', True))

    @property
    def token_expiration(self):
        """ The amount of time an e-mail login token is valid for. """
        expires = self._settings.get('email_token_expiration_hours', 3)
        return timedelta(hours=expires)

    @property
    def default_url(self):
        """ Default redirect URL after login if ``next`` is not provided. This is just a standard Django setting. """
        return django_settings.LOGIN_REDIRECT_URL  # Can be assumed to be set.


settings = AppSettings()



# -*- coding: utf-8 -*-
from logging import getLogger
from threading import Thread

from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required
from django.core.exceptions import SuspiciousOperation
from django.core.mail import EmailMessage
from django.core.urlresolvers import reverse
from django.http.response import HttpResponseRedirectBase
from django.middleware.csrf import CsrfViewMiddleware
from django.shortcuts import render, resolve_url
from django.utils.translation import ugettext_lazy as _t, ugettext as _tt
from django.views.decorators.csrf import csrf_exempt

from .app import settings
from .auth import OneAllAuthBackend, EmailTokenAuthBackend
from .forms import EmailForm
from .models import SocialUserCache

log = getLogger(__name__)


class HttpResponseSeeOther(HttpResponseRedirectBase):
    status_code = 303


def redirect(to, *args, **kwargs):
    return HttpResponseSeeOther(resolve_url(to, *args, **kwargs))


@csrf_exempt
def oa_login(request, noise=''):
    """ Display and callback view for OneAll Authentication. """
    if request.user and request.user.is_authenticated():
        return redirect('oneall-profile')
    auth_with = None
    context = {
        'login_failed': False,
        'check_your_mail': False,
        'logged_out': 'logout' in noise,
        'email_form': EmailForm(),
    }
    if OneAllAuthBackend.KEY in request.POST:
        auth_with = dict(request.POST.items())
    elif 'email' in request.POST:
        csrf_check(request)  # maybe not relevant
        email = request.POST['email']
        args = EmailTokenAuthBackend().issue(email)
        mail_login_token(request, email, args)
        context['check_your_mail'] = True
    elif EmailTokenAuthBackend.KEY in request.GET:
        auth_with = dict(request.GET.items())
    if auth_with:
        user = authenticate(**auth_with)
        if user:
            login(request, user)
            return redirect(request.GET.get('next') or settings.default_url)
        else:
            context['login_failed'] = True
    return render(request, 'oneall/login.html', context)


def mail_login_token(request, email, args):
    args.setlist('next', request.GET.getlist('next'))
    relative_uri = '%s?%s' % (reverse('oneall-login'), args.urlencode())
    message = EmailMessage()
    message.subject = _t("Login")
    message.to = [email]
    message.body = "\n".join([
        _tt("Complete your login using this link:"),
        request.build_absolute_uri(relative_uri),
    ])
    Thread(target=message.send, args=(True,)).start()


def oa_logout(request):
    """ Logs out the user and then takes them somewhere else. """
    logout(request)
    url = request.GET.get('next')
    if url:
        return redirect(url)
    else:
        return redirect('oneall-login', '_from_logout')


@login_required
@csrf_exempt
def oa_profile(request):
    """ Displays current user profile, allows updates and social linkings. """
    context = {
        'user': request.user,
        'identity': SocialUserCache.objects.filter(user=request.user).first(),
        'form': EmailForm({'email': request.user.email})
    }
    if 'connection_token' in request.POST:
        OneAllAuthBackend(request.user).authenticate(**request.POST)
    elif 'email' in request.POST and request.user.email != request.POST['email']:
        csrf_check(request)
        EmailTokenAuthBackend(request.user).issue(request.POST['email'])
    return render(request, 'oneall/profile.html', context)


def csrf_check(request, raise_exception=True):
    problem = CsrfViewMiddleware().process_view(request, None, (), {})
    if problem and raise_exception:
        raise SuspiciousOperation
    else:
        return not problem
